﻿using PalmSens.Core.Simplified.Data;
using PalmSens.Core.UniversalWindows;
using PalmSens.Data;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Storage;

namespace PalmSens.Core.Simplified.UniversalWindows
{
    public class SimpleLoadSaveFunctions
    {
        /// <summary>
        /// Loads a collection of simplemeasurements from a *.pssession file.
        /// </summary>
        /// <param name="storageFile">The storage file.</param>
        /// <returns>
        /// Collection of simplemeasurements
        /// </returns>
        /// <exception cref="ArgumentNullException">The target storage file must be specified.</exception>
        /// <exception cref="Exception">An error occured while loading, please make sure the file path is correct and the file is valid</exception>
        public static async Task<List<SimpleMeasurement>> LoadMeasurements(StorageFile storageFile)
        {
            if (storageFile == null)
                throw new ArgumentNullException("The target storage file must be specified.");

            List<SimpleMeasurement> simpleMeasurements = new List<SimpleMeasurement>();
            SessionManager session = null;

            try { session = await LoadSaveHelperFunctions.LoadSessionFile(storageFile); }
            catch (Exception ex)
            {
                throw new Exception("An error occured while loading, please make sure the file path is correct and the file is valid");
            }

            if (session != null)
                foreach (Measurement measurement in session)
                    simpleMeasurements.Add(new SimpleMeasurement(measurement));

            return simpleMeasurements;
        }

        /// <summary>
        /// Saves a simplemeasurement to a *.pssession file.
        /// </summary>
        /// <param name="simpleMeasurement">The simple measurement.</param>
        /// <param name="storageFile">The storage file.</param>
        /// <returns></returns>
        /// <exception cref="ArgumentNullException">
        /// The target storage file must be specified.
        /// or
        /// SimpleMeasurement cannot be null
        /// </exception>
        /// <exception cref="Exception">An error occured while saving, please make sure the file path is correct</exception>
        public static async Task SaveMeasurement(SimpleMeasurement simpleMeasurement, StorageFile storageFile)
        {
            if (storageFile == null)
                throw new ArgumentNullException("The target storage file must be specified.");
            if (simpleMeasurement == null)
                throw new ArgumentNullException("SimpleMeasurement cannot be null");

            SessionManager session = new SessionManager();
            session.AddMeasurement(simpleMeasurement.Measurement);
            session.MethodForEditor = simpleMeasurement.Measurement.Method;

            try { await LoadSaveHelperFunctions.SaveSessionFile(session, storageFile); }
            catch (Exception ex)
            {
                throw new Exception("An error occured while saving, please make sure the file path is correct");
            }
        }

        /// <summary>
        /// Saves a collection of  simplemeasurements to a *.pssession file.
        /// </summary>
        /// <param name="simpleMeasurements">Array of simplemeasurements.</param>
        /// <param name="storageFile">The storage file.</param>
        /// <returns></returns>
        /// <exception cref="ArgumentNullException">
        /// The target storage file must be specified.
        /// or
        /// SimpleMeasurements cannot be null
        /// </exception>
        /// <exception cref="Exception">An error occured while saving, please make sure the file path is correct</exception>
        public static async Task SaveMeasurements(SimpleMeasurement[] simpleMeasurements, StorageFile storageFile)
        {
            if (storageFile == null)
                throw new ArgumentNullException("The target storage file must be specified.");
            if (simpleMeasurements == null || simpleMeasurements.Where(meas => meas == null).Count() > 0)
                throw new ArgumentNullException("SimpleMeasurements cannot be null");

            SessionManager session = new SessionManager();
            foreach (SimpleMeasurement measurement in simpleMeasurements)
                if (measurement != null)
                    session.AddMeasurement(measurement.Measurement);
            session.MethodForEditor = simpleMeasurements[0].Measurement.Method;

            try { await LoadSaveHelperFunctions.SaveSessionFile(session, storageFile); }
            catch (Exception ex)
            {
                throw new Exception("An error occured while saving, please make sure the file path is correct");
            }
        }

        /// <summary>
        /// Saves a collection of  simplemeasurements to a *.pssession file.
        /// </summary>
        /// <param name="simpleMeasurements">Array of simplemeasurements.</param>
        /// <param name="storageFile">The storage file.</param>
        /// <returns></returns>
        /// <exception cref="ArgumentNullException">
        /// The target storage file must be specified.
        /// or
        /// SimpleMeasurements cannot be null
        /// </exception>
        public static async Task SaveMeasurements(List<SimpleMeasurement> simpleMeasurements, StorageFile storageFile)
        {
            if (storageFile == null)
                throw new ArgumentNullException("The target storage file must be specified.");
            if (simpleMeasurements == null || simpleMeasurements.Where(meas => meas == null).Count() > 0)
                throw new ArgumentNullException("SimpleMeasurements cannot be null");

            await SaveMeasurements(simpleMeasurements.ToArray(), storageFile);
        }

        /// <summary>
        /// Loads a method from a *.psmethod file.
        /// </summary>
        /// <param name="storageFile">The storage file.</param>
        /// <returns>
        /// Method
        /// </returns>
        /// <exception cref="ArgumentNullException">The target storage file must be specified.</exception>
        /// <exception cref="Exception">An error occured while loading, please make sure the file path is correct and the file is valid</exception>
        public static async Task<Method> LoadMethod(StorageFile storageFile)
        {
            Method method = null;
            if (storageFile == null)
                throw new ArgumentNullException("The target storage file must be specified.");

            try { method = await LoadSaveHelperFunctions.LoadMethod(storageFile); }
            catch (Exception ex)
            {
                throw new Exception("An error occured while loading, please make sure the file path is correct and the file is valid");
            }

            return method;
        }

        /// <summary>
        /// Saves a method to a *.psmethod file.
        /// </summary>
        /// <param name="method">The method .</param>
        public static async Task SaveMethod(Method method, StorageFile storageFile)
        {
            if (storageFile == null)
                throw new ArgumentNullException("The target storage file must be specified.");
            if (method == null)
                throw new ArgumentNullException("Method cannot be null");

            try { await LoadSaveHelperFunctions.SaveMethod(method, storageFile); }
            catch (Exception ex)
            {
                throw new Exception("An error occured while saving, please make sure the file path is correct");
            }
        }
    }
}
