﻿using PalmSens;
using PalmSens.Comm;
using PalmSens.Core.Simplified.Data;
using PalmSens.Devices;
using PalmSens.Techniques;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using System.Threading.Tasks;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=402352&clcid=0x409

namespace PSSDKMultiChannelExample
{
    /// <summary>
    /// An empty page that can be used on its own or navigated to within a Frame.
    /// </summary>
    public sealed partial class MainPage : Page
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="MainPage"/> class.
        /// </summary>
        public MainPage()
        {
            this.InitializeComponent();
        }

        /// <summary>
        /// Handles the Loaded event of the Page control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private void Page_Loaded(object sender, RoutedEventArgs e)
        {
            InitCVMethod(); //Create the cyclic voltammetry method that defines the measurement parameters
            InitPlot(); //Resets and initiates the plot control
            lbDevices.ItemsSource = _connectedDevices;
            DiscoverConnectedDevices(); //Populate the connected device combobox control
        }

        /// <summary>
        /// The instance of method class containing the Cyclic Voltammetry parameters
        /// </summary>
        private CyclicVoltammetry _methodCV;

        /// <summary>
        /// The active SimpleMeasurement
        /// </summary>
        private SimpleMeasurement[] _activeMeasurement = null;

        /// <summary>
        /// The connected PalmSens & EmStat devices
        /// </summary>
        private ObservableCollection<DeviceItem> _connectedDevices = new ObservableCollection<DeviceItem>();

        /// <summary>
        /// Gets the selected channels.
        /// </summary>
        /// <value>
        /// The selected channels.
        /// </value>
        private int[] SelectedChannels
        {
            get
            {
                List<int> selectedChannels = new List<int>();

                for (int i = 0; i < _connectedDevices.Count; i++)
                    if (_connectedDevices[i].Checked)
                        selectedChannels.Add(i);

                return selectedChannels.ToArray();
            }
        }

        /// <summary>
        /// Gets the selected devices.
        /// </summary>
        /// <value>
        /// The selected devices.
        /// </value>
        private Device[] SelectedDevices
        {
            get
            {
                int[] selectedChannels = SelectedChannels;
                int n = selectedChannels.Length;
                Device[] devices = new Device[n];
                for (int i = 0; i < n; i++)
                    devices[i] = _connectedDevices[selectedChannels[i]].Device;
                return devices;
            }
        }

        /// <summary>
        /// Initializes the CV method.
        /// </summary>
        private void InitCVMethod()
        {
            _methodCV = new CyclicVoltammetry(); //Create a new cyclic voltammetry method with the default settings
            _methodCV.BeginPotential = -.5f; //Sets the potential to start the scan from
            _methodCV.Vtx1Potential = -.5f; //Sets the first potential where the scan direction reverses
            _methodCV.Vtx2Potential = .5f; //Sets the second potential where the scan direction reverses
            _methodCV.StepPotential = 0.01f; //Sets the step size
            _methodCV.Scanrate = 1f; //Sets the scan rate to 1 V/s
            _methodCV.nScans = 3; //Sets the number of scans

            _methodCV.EquilibrationTime = 1f; //Equilabrates the cell at the defined potential for 1 second before starting the measurement
            _methodCV.Ranging.StartCurrentRange = new CurrentRange(CurrentRanges.cr1uA); //Starts equilabration in the 1µA current range
            _methodCV.Ranging.MinimumCurrentRange = new CurrentRange(CurrentRanges.cr10nA); //Min current range 10nA
            _methodCV.Ranging.MaximumCurrentRange = new CurrentRange(CurrentRanges.cr1mA); //Max current range 1mA
        }

        /// <summary>
        /// Initializes the plot control.
        /// </summary>
        private void InitPlot()
        {
            plot.ClearAll(); //Clear all curves and data from plot
            //Set the Axis labels
            plot.XAxisLabel = "V";
            plot.YAxisLabel = "µA";
            plot.AddData("", new double[0], new double[0]); //Add a empty data array to draw an empty plot
        }

        /// <summary>
        /// Discovers the connected PalmSens & EmStat devices and adds them to the combobox control.
        /// </summary>
        private async Task DiscoverConnectedDevices()
        {
            btnRefresh.IsEnabled = false;
            _connectedDevices.Clear();
            Device[] devices = await psMultiCommSimpleUWP.GetConnectedDevicesAsync(); //Discover connected devices

            foreach (Device d in devices)
                _connectedDevices.Add(new DeviceItem() { Checked = false, Device = d, Title = d.ToString() }); //Add connected devices to control

            int nDevices = _connectedDevices.Count;
            lbConsole.Items.Add($"Found {nDevices} device(s).");
            btnConnect.IsEnabled = nDevices > 0;
            btnRefresh.IsEnabled = true;
        }

        /// <summary>
        /// Handles the Click event of the btnRefresh control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnRefresh_Click(object sender, RoutedEventArgs e)
        {
            await DiscoverConnectedDevices();
        }

        /// <summary>
        /// Handles the Click event of the btnConnect control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnConnect_Click(object sender, RoutedEventArgs e)
        {
            btnConnect.IsEnabled = false;
            if (!psMultiCommSimpleUWP.Connected) //Determine whether a device is currently connected
            {
                if (SelectedChannels.Length == 0)
                {
                    btnConnect.IsEnabled = true;
                    return;
                }

                try
                {
                    //Connect to the device selected in the devices listbox control
                    await psMultiCommSimpleUWP.ConnectAsync(SelectedDevices);
                    lbConsole.Items.Add($"Connected to {psMultiCommSimpleUWP.NConnectedChannels} channels");
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                await psMultiCommSimpleUWP.Disconnect(); //Disconnect from the connected device
                lbConsole.Items.Add("Disconnected");
                await DiscoverConnectedDevices();
            }

            //Update UI based on connection status
            if (psMultiCommSimpleUWP.Connected)
            {
                //Updated selected channel list and connected device combobox
                _connectedDevices.Clear();
                cmbDevices.Items.Clear();
                for (int i = 0; i < psMultiCommSimpleUWP.NConnectedChannels; i++)
                {
                    string chnl = $"{i + 1}: {psMultiCommSimpleUWP.ConnectedDevices[i].ToString()}";
                    _connectedDevices.Add(new DeviceItem() { Checked = false, Device = psMultiCommSimpleUWP.ConnectedDevices[i], Title = chnl });
                    cmbDevices.Items.Add(chnl);
                }
                cmbDevices.SelectedIndex = 0;
            }

            lblDevices.Text = psMultiCommSimpleUWP.Connected ? "Connected channels:" : "Available devices";
            cmbDevices.IsEnabled = psMultiCommSimpleUWP.Connected;
            btnRefresh.IsEnabled = !psMultiCommSimpleUWP.Connected;
            btnConnect.Content = psMultiCommSimpleUWP.Connected ? "Disconnect" : "Connect";
            btnMeasure.IsEnabled = psMultiCommSimpleUWP.Connected;
            btnConnect.IsEnabled = false;
        }

        /// <summary>
        /// Handles the Click event of the btnMeasure control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="RoutedEventArgs"/> instance containing the event data.</param>
        private async void btnMeasure_Click(object sender, RoutedEventArgs e)
        {
            int[] selectedChannels = SelectedChannels;
            if (selectedChannels.Length == 0) return;

            btnMeasure.IsEnabled = false;
            lbDevices.IsEnabled = false;

            if (selectedChannels.Where(c => psMultiCommSimpleUWP.ChannelStates[c] == PalmSens.Comm.CommManager.DeviceState.Idle).Count() == selectedChannels.Length) //Determine whether the devices are currently idle or measuring
            {
                try
                {
                    plot.ClearAll(); //Clears data from previous measurements from the plot
                    _activeMeasurement = await psMultiCommSimpleUWP.MeasureAsync(_methodCV, selectedChannels); //Start measurement defined in the method
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            else
            {
                try
                {
                    await psMultiCommSimpleUWP.AbortAllActiveMeasurementsAsync(); //Abort the active measurement
                }
                catch (Exception ex)
                {
                    lbConsole.Items.Add(ex.Message);
                }
            }
            btnMeasure.IsEnabled = true;
        }

        /// <summary>
        /// Raised when device status package is received (the device does not send status packages while measuring)
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="status">The <see cref="StatusEventArgs"/> instance containing the event data.</param>
        /// <param name="channel">The channel.</param>
        private void PsMultiCommSimpleUWP_ReceiveStatus(object sender, StatusEventArgs e, int channel)
        {
            if (cmbDevices.SelectedIndex == channel)
            {
                Status status = e.GetStatus(); //Get the PalmSens.Comm.Status instance from the event data
                double potential = status.PotentialReading.Value; //Get the potential
                double currentInRange = status.CurrentReading.ValueInRange; //Get the current expressed inthe active current range
                PalmSens.Comm.ReadingStatus currentStatus = status.CurrentReading.ReadingStatus; //Get the status of the current reading
                CurrentRange cr = status.CurrentReading.CurrentRange; //Get the active current range

                tbPotential.Text = potential.ToString("F3");
                tbCurrent.Text = currentInRange.ToString("F3");
                lblCurrentRange.Text = $"* {cr.ToString()}";
            }
        }

        /// <summary>
        /// Raised when the connected device's status changes
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="ChannelState">State of the channel.</param>
        /// <param name="channel">The channel.</param>
        private void PsMultiCommSimpleUWP_StateChanged(object sender, CommManager.DeviceState ChannelState, int channel)
        {
            if (cmbDevices.SelectedIndex == channel)
            {
                tbStatus.Text = ChannelState.ToString(); //Updates the device state indicator textbox
                btnConnect.IsEnabled = ChannelState == PalmSens.Comm.CommManager.DeviceState.Idle;
                btnMeasure.Content = ChannelState == PalmSens.Comm.CommManager.DeviceState.Idle ? "Measure" : "Abort";
            }
        }

        /// <summary>
        /// Raised when the measurement is ended
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="channel">The channel.</param>
        private void PsMultiCommSimpleUWP_MeasurementStarted(object sender, int channel)
        {
            lbConsole.Items.Add($"Channel {channel + 1}: Cyclic voltammetry measurement started.");
        }

        /// <summary>
        /// Raised when the measurement is started
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="channel">The channel.</param>
        private void PsMultiCommSimpleUWP_MeasurementEnded(object sender, int channel)
        {
            lbConsole.Items.Add($"Channel {channel + 1}: Measurement ended.");
            lbDevices.IsEnabled = psMultiCommSimpleUWP.ChannelStates.Where(c => c == CommManager.DeviceState.Idle).Count() == psMultiCommSimpleUWP.NConnectedChannels;
        }

        /// <summary>
        /// Raised when a Simple Curve in the active SimpleMeasurement starts receiving data
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="activeSimpleCurve">The active simple curve.</param>
        private void PsMultiCommSimpleUWP_SimpleCurveStartReceivingData(object sender, SimpleCurve activeSimpleCurve)
        {
            plot.AddSimpleCurve(activeSimpleCurve);

            //Subscribe to the event indicating when the curve stops receiving new data points
            activeSimpleCurve.CurveFinished += activeSimpleCurve_CurveFinished;

            lbConsole.Items.Add("Curve is receiving new data...");
        }

        /// <summary>
        /// Raised when a SimpleCurve stops receiving new data points
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void activeSimpleCurve_CurveFinished(object sender, EventArgs e)
        {
            if (!Dispatcher.HasThreadAccess) //Data is parsed asynchronously in the case this event was raised on a different thread it must be invoked back onto the UI thread
            {
                Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () => { activeSimpleCurve_CurveFinished(sender, e); });
                return;
            }
            SimpleCurve activeCurve = sender as SimpleCurve;
            int nDataPointsReceived = activeCurve != null ? activeCurve.NDataPoints : 0;
            lbConsole.Items.Add($"{nDataPointsReceived} data point(s) received.");

            //Unsubscribe from the curves events to avoid memory leaks
            activeCurve.CurveFinished -= activeSimpleCurve_CurveFinished;
        }
    }

    /// <summary>
    /// Object class for available/connected devices in lbDevices with checkbox and label
    /// </summary>
    public class DeviceItem
    {
        public bool Checked { get; set; }
        public string Title { get; set; }
        public Device Device { get; set; }
    }
}
