﻿using PalmSens.Comm;
using PalmSens.Core.Simplified;
using PalmSens.Core.Simplified.Data;
using PalmSens.Devices;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace PalmSens.Core.Simplified.UniversalWindows
{
    public sealed partial class PSMultiCommSimpleUWP : UserControl, IPlatformMulti, IDisposable
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="PSMultiCommSimpleUWP"/> class.
        /// This class handles the communication with a device and it is used
        /// to perform measurements and control the device manually.
        /// </summary>
        public PSMultiCommSimpleUWP()
        {
            InitializeComponent();
            PalmSens.UniversalWindows.CoreDependencies.Init(Application.Current); //Initiates PSSDK threading dependencies
            InitAsyncFunctionality(Environment.ProcessorCount); //Initiate the asynchronous functions in the SDK
            _psMultiCommSimple = new PSMultiCommSimple(this);
        }

        #region Properties
        /// <summary>
        /// Instance of the platform independent PSCommSimple class that manages measurements and manual control
        /// </summary>
        private PSMultiCommSimple _psMultiCommSimple;

        /// <summary>
        /// Gets the CommManager for the current connection.
        /// </summary>
        /// <value>
        /// The CommManager, null when there is no active connection to a channel.
        /// </value>
        public CommManager[] Comms { get { return _psMultiCommSimple.Comms; } }

        /// <summary>
        /// The device handler class which handles the connection to the device
        /// </summary>
        private DeviceHandler _deviceHandler = new DeviceHandler();

        /// <summary>
        /// Gets a value indicating whether <see cref="PSCommSimple"/> is connected to a channel.
        /// </summary>
        /// <value>
        ///   <c>true</c> if connected; otherwise, <c>false</c>.
        /// </value>
        public bool Connected { get { return _psMultiCommSimple.Connected; } }

        /// <summary>
        /// Gets the number of connected channels.
        /// </summary>
        /// <value>
        /// The n connected channels.
        /// </value>
        public int NConnectedChannels { get { return _psMultiCommSimple.NConnectedChannels; } }

        /// <summary>
        /// Gets the connected channel type.
        /// </summary>
        /// <value>
        /// The connected channel type.
        /// </value>
        public enumDeviceType[] ConnectedChannels { get { return _psMultiCommSimple.ConnectedChannels; } }

        /// <summary>
        /// Returns an array of all of the connected devices.
        /// </summary>
        /// <value>
        /// The connected channels.
        /// </value>
        public Device[] ConnectedDevices { get { return _psMultiCommSimple.ConnectedDevices; } }

        /// <summary>
        /// Gets the state of the connected channel.
        /// </summary>
        /// <value>
        /// The state of the channel.
        /// </value>
        public CommManager.DeviceState[] ChannelStates { get { return _psMultiCommSimple.ChannelStates; } }

        /// <summary>
        /// Gets a value indicating whether [cell is on].
        /// </summary>
        /// <value>
        ///   <c>true</c> if [cell is on]; otherwise, <c>false</c>.
        /// </value>
        public bool[] CellOn { get { return _psMultiCommSimple.IsCellOn; } }

        /// <summary>
        /// Gets the capabilities of the connected channel.
        /// </summary>
        /// <value>
        /// The channel capabilities.
        /// </value>
        public DeviceCapabilities[] Capabilities { get { return _psMultiCommSimple.Capabilities; } }

        /// <summary>
        /// Determines whether [the specified method] is compatible with all connected channels.
        /// </summary>
        /// <param name="method">The method.</param>
        /// <returns>
        ///   <c>true</c> if the method is valid; otherwise, <c>false</c>.
        /// </returns>
        public bool[] IsValidMethod(Method method)
        {
            return _psMultiCommSimple.IsValidMethod(method);
        }

        /// <summary>
        /// Determines whether [the specified method] is compatible with the specified channel.
        /// </summary>
        /// <param name="method">The method.</param>
        /// <param name="channel">The specified channel.</param>
        /// <returns>
        ///   <c>true</c> if the method is valid; otherwise, <c>false</c>.
        /// </returns>
        public bool IsValidMethod(Method method, int channel)
        {
            return _psMultiCommSimple.IsValidMethod(method, channel);
        }

        /// <summary>
        /// Determines whether [the specified method] is compatible with the specified channels.
        /// </summary>
        /// <param name="method">The method.</param>
        /// <param name="channels">The specified channels.</param>
        /// <returns>
        ///   <c>true</c> if the method is valid; otherwise, <c>false</c>.
        /// </returns>
        public bool[] IsValidMethod(Method method, int[] channels)
        {
            return _psMultiCommSimple.IsValidMethod(method, channels);
        }
        #endregion

        #region Functions     
        /// <summary>
        /// Required initialization for using the async functionalities of the PalmSens SDK.
        /// The amount of simultaneous operations will be limited to prevent performance issues.
        /// When possible it will leave one core free for the UI.
        /// </summary>
        /// <param name="nCores">The number of CPU cores.</param>
        public void InitAsyncFunctionality(int nCores)
        {
            SynchronizationContextRemover.Init(nCores > 1 ? nCores - 1 : 1);
        }

        /// <summary>
        /// Returns an array of connected channels.
        /// </summary>
        public async Task<Device[]> GetConnectedDevicesAsync()
        {
            return await _deviceHandler.GetConnectedDevicesAsync();
        }

        /// <summary>
        /// Connects to the specified channel.
        /// </summary>
        /// <param name="device">The device.</param>
        public async Task ConnectAsync(Device[] devices)
        {
            _psMultiCommSimple.Comms = await _deviceHandler.ConnectAsync(devices);
        }

        /// <summary>
        /// Disconnects from the connected channel.
        /// </summary>
        public async Task Disconnect()
        {
            await _psMultiCommSimple.Disconnect();
        }

        /// <summary>
        /// Turns the cell on on specified channel.
        /// </summary>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task TurnCellOnAsync(int channel)
        {
            await _psMultiCommSimple.TurnCellOnAsync(channel);
        }

        /// <summary>
        /// Turns the cell on on specified channels.
        /// </summary>
        /// <param name="channels">The channels.</param>
        /// <returns></returns>
        public async Task TurnCellOnAsync(int[] channels)
        {
            await _psMultiCommSimple.TurnCellOnAsync(channels);
        }

        /// <summary>
        /// Turns the cell on on all channels.
        /// </summary>
        public async Task TurnCellOnAsync()
        {
            await _psMultiCommSimple.TurnCellOnAsync();
        }

        /// <summary>
        /// Turns the cell off on specified channel.
        /// </summary>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task TurnCellOffAsync(int channel)
        {
            await _psMultiCommSimple.TurnCellOffAsync(channel);
        }

        /// <summary>
        /// Turns the cell off on specified channels.
        /// </summary>
        /// <param name="channels">The channels.</param>
        /// <returns></returns>
        public async Task TurnCellOffAsync(int[] channels)
        {
            await _psMultiCommSimple.TurnCellOffAsync(channels);
        }

        /// <summary>
        /// Turns the cell off on all channels.
        /// </summary>
        public async Task TurnCellOffAsync()
        {
            await _psMultiCommSimple.TurnCellOffAsync();
        }

        /// <summary>
        /// Sets the cell potential on the specified channel.
        /// </summary>
        /// <param name="potential">The potential.</param>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task SetCellPotentialAsync(float potential, int channel)
        {
            await _psMultiCommSimple.SetCellPotentialAsync(potential, channel);
        }

        /// <summary>
        /// Sets the cell potential on the specified channels.
        /// </summary>
        /// <param name="potential">The potential.</param>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task SetCellPotentialAsync(float potential, int[] channels)
        {
            await _psMultiCommSimple.SetCellPotentialAsync(potential, channels);
        }

        /// <summary>
        /// Sets the cell potential on all channels.
        /// </summary>
        /// <param name="potential">The potential.</param>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task SetCellPotentialAsync(float potential)
        {
            await _psMultiCommSimple.SetCellPotentialAsync(potential);
        }

        /// <summary>
        /// Reads the cell potential on the specified channel.
        /// </summary>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task<float> ReadCellPotentialAsync(int channel)
        {
            return await _psMultiCommSimple.ReadCellPotentialAsync(channel);
        }

        /// <summary>
        /// Reads the cell potential on the specified channels.
        /// </summary>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task<float[]> ReadCellPotentialAsync(int[] channels)
        {
            return await _psMultiCommSimple.ReadCellPotentialAsync(channels);
        }

        /// <summary>
        /// Reads the cell potential on all channels.
        /// </summary>
        /// <returns></returns>
        public async Task<float[]> ReadCellPotentialAsync()
        {
            return await _psMultiCommSimple.ReadCellPotentialAsync();
        }

        /// <summary>
        /// Sets the cell current on the specified channel.
        /// </summary>
        /// <param name="current">The current.</param>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task SetCellCurrentAsync(float current, int channel)
        {
            await _psMultiCommSimple.SetCellCurrentAsync(current, channel);
        }

        /// <summary>
        /// Sets the cell current on the specified channels.
        /// </summary>
        /// <param name="current">The current.</param>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task SetCellCurrentAsync(float current, int[] channels)
        {
            await _psMultiCommSimple.SetCellCurrentAsync(current, channels);
        }

        /// <summary>
        /// Sets the cell current on all channels.
        /// </summary>
        /// <param name="current">The current.</param>
        /// <param name="channel">The channel.</param>
        /// <returns></returns>
        public async Task SetCellCurrentAsync(float current)
        {
            await _psMultiCommSimple.SetCellCurrentAsync(current);
        }

        /// <summary>
        /// Reads the cell current on the specified channel.
        /// </summary>
        /// <returns></returns>
        public async Task<float> ReadCellCurrentAsync(int channel)
        {
            return await _psMultiCommSimple.ReadCellCurrentAsync(channel);
        }

        /// <summary>
        /// Reads the cell current on the specified channels.
        /// </summary>
        /// <returns></returns>
        public async Task<float[]> ReadCellCurrentAsync(int[] channels)
        {
            return await _psMultiCommSimple.ReadCellCurrentAsync(channels);
        }

        /// <summary>
        /// Reads the cell current on all channels.
        /// </summary>
        /// <returns></returns>
        public async Task<float[]> ReadCellCurrentAsync()
        {
            return await _psMultiCommSimple.ReadCellCurrentAsync();
        }

        /// <summary>
        /// Sets the current range on the specified channel.
        /// </summary>
        /// <param name="currentRange">The current range.</param>
        public async Task SetCurrentRangeAsync(CurrentRange currentRange, int channel)
        {
            await _psMultiCommSimple.SetCurrentRangeAsync(currentRange, channel);
        }

        /// <summary>
        /// Sets the current range on the specified channels.
        /// </summary>
        /// <param name="currentRange">The current range.</param>
        public async Task SetCurrentRangeAsync(CurrentRange currentRange, int[] channels)
        {
            await _psMultiCommSimple.SetCurrentRangeAsync(currentRange, channels);
        }

        /// <summary>
        /// Sets the current range on all channels.
        /// </summary>
        /// <param name="currentRange">The current range.</param>
        public async Task SetCurrentRangeAsync(CurrentRange currentRange)
        {
            await _psMultiCommSimple.SetCurrentRangeAsync(currentRange);
        }

        /// <summary>
        /// Runs a measurement as specified in the method on the specified channel.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <param name="channel">The channel.</param>
        /// <param name="muxChannel">The mux channel to measure on.</param>
        /// <returns>
        /// A SimpleMeasurement instance containing all the data related to the measurement.
        /// </returns>
        public async Task<SimpleMeasurement> MeasureAsync(Method method, int channel, int muxChannel, TaskBarrier taskBarrier = null)
        {
            return await _psMultiCommSimple.MeasureAsync(method, channel, muxChannel);
        }

        /// <summary>
        /// Runs a measurement as specified in the method on the specified collection of channels.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <param name="channels">The channels.</param>
        /// <param name="muxChannel">The mux channel to measure on.</param>
        /// <returns>
        /// A SimpleMeasurement instance containing all the data related to the measurement.
        /// </returns>
        public async Task<SimpleMeasurement[]> MeasureAsync(Method method, int[] channels, int muxChannel)
        {
            return await _psMultiCommSimple.MeasureAsync(method, channels, muxChannel);
        }

        /// <summary>
        /// Runs a measurement as specified in the method on all channels.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <param name="channels">The channels.</param>
        /// <param name="muxChannel">The mux channel to measure on.</param>
        /// <returns>
        /// A SimpleMeasurement instance containing all the data related to the measurement.
        /// </returns>
        public async Task<SimpleMeasurement[]> MeasureAllChannelsAsync(Method method, int muxChannel)
        {
            return await _psMultiCommSimple.MeasureAllChannelsAsync(method, muxChannel);
        }

        /// <summary>
        /// Runs a measurement as specified in the method on the specified channel.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <returns>A SimpleMeasurement instance containing all the data related to the measurement.</returns>
        public async Task<SimpleMeasurement> MeasureAsync(Method method, int channel)
        {
            return await _psMultiCommSimple.MeasureAsync(method, channel);
        }

        /// <summary>
        /// Runs a measurement as specified in the method on the specified channel.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <param name="channels">The channels.</param>
        /// <returns>
        /// A SimpleMeasurement instance containing all the data related to the measurement.
        /// </returns>
        public async Task<SimpleMeasurement[]> MeasureAsync(Method method, int[] channels)
        {
            return await _psMultiCommSimple.MeasureAsync(method, channels);
        }

        /// <summary>
        /// Runs a measurement as specified in the method on the specified channel.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <param name="channels">The channels.</param>
        /// <returns>
        /// A SimpleMeasurement instance containing all the data related to the measurement.
        /// </returns>
        public async Task<SimpleMeasurement[]> MeasureAllChannelsAsync(Method method)
        {
            return await _psMultiCommSimple.MeasureAllChannelsAsync(method);
        }

        /// <summary>
        /// Aborts the active measurement on the specified channel.
        /// </summary>
        public async Task AbortMeasurementAsync(int channel)
        {
            await _psMultiCommSimple.AbortMeasurementAsync(channel);
        }

        /// <summary>
        /// Aborts the active measurement on the specified channels.
        /// </summary>
        public async Task AbortMeasurementsAsync(int[] channels)
        {
            await _psMultiCommSimple.AbortMeasurementsAsync(channels);
        }

        /// <summary>
        /// Aborts all active measurements channels.
        /// </summary>
        public async Task AbortAllActiveMeasurementsAsync()
        {
            await _psMultiCommSimple.AbortAllActiveMeasurementsAsync();
        }

        /// <summary>
        /// Validates whether the specified method is compatible with the capabilities of the specified connected channel.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <param name="channel">The specified channel.</param>
        /// <param name="isValidMethod">if set to <c>true</c> [is valid method].</param>
        /// <param name="errors">The errors.</param>
        public void ValidateMethod(Method method, int channel, out bool isValidMethod, out List<string> errors)
        {
            _psMultiCommSimple.ValidateMethod(method, channel, out isValidMethod, out errors);
        }

        /// <summary>
        /// Validates whether the specified method is compatible with the capabilities of the specified connected channels.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <param name="channels">The specified channels.</param>
        /// <param name="isValidMethod">if set to <c>true</c> [is valid method].</param>
        /// <param name="errors">The errors.</param>
        public void ValidateMethod(Method method, int[] channels, out bool[] isValidMethod, out List<string>[] errors)
        {
            _psMultiCommSimple.ValidateMethod(method, channels, out isValidMethod, out errors);
        }

        /// <summary>
        /// Validates whether the specified method is compatible with the capabilities of all connected channels.
        /// </summary>
        /// <param name="method">The method containing the measurement parameters.</param>
        /// <param name="isValidMethod">if set to <c>true</c> [is valid method].</param>
        /// <param name="errors">The errors.</param>
        public void ValidateMethod(Method method, out bool[] isValidMethod, out List<string>[] errors)
        {
            _psMultiCommSimple.ValidateMethod(method, out isValidMethod, out errors);
        }
        #endregion

        #region Platform interface
        /// <summary>
        /// Invokes event to UI thread if required.
        /// </summary>
        /// <param name="method">The method.</param>
        /// <param name="args">The arguments.</param>
        /// <returns></returns>
        /// <exception cref="System.NullReferenceException">Parent control not set.</exception>
        public bool InvokeIfRequired(Delegate method, params object[] args)
        {
            if (!Dispatcher.HasThreadAccess) //Check if event needs to be cast to the UI thread
            {
                Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () => method.DynamicInvoke(args)); //Recast event to UI thread
                return true;
            }
            return false;
        }

        /// <summary>
        /// Disconnects from device with the specified CommManagers. 
        /// Warning use the platform independent method Disconnect() instead.
        /// Otherwise the generic PSCommSimple does not unsubscribe from the CommManager correctly
        /// which may result in it not being released from the memory.
        /// </summary>
        /// <param name="comm">The comm.</param>
        public async Task Disconnect(CommManager[] comms)
        {
            await _deviceHandler.Disconnect(comms);
        }
        #endregion

        #region events
        /// <summary>
        /// Occurs when a channel status package is received, these packages are not sent during a measurement.
        /// </summary>
        public event MultiChannelStatusEventHandler ReceiveStatus
        {
            add { _psMultiCommSimple.ReceiveStatus += value; }
            remove { _psMultiCommSimple.ReceiveStatus -= value; }
        }

        /// <summary>
        /// Occurs at the start of a new measurement.
        /// </summary>
        public event MultiChannelEventHandler MeasurementStarted
        {
            add { _psMultiCommSimple.MeasurementStarted += value; }
            remove { _psMultiCommSimple.MeasurementStarted -= value; }
        }

        /// <summary>
        /// Occurs when a measurement has ended.
        /// </summary>
        public event MultiChannelEventHandler MeasurementEnded
        {
            add { _psMultiCommSimple.MeasurementEnded += value; }
            remove { _psMultiCommSimple.MeasurementEnded -= value; }
        }

        /// <summary>
        /// Occurs when a new [SimpleCurve starts receiving data].
        /// </summary>
        public event PSCommSimple.SimpleCurveStartReceivingDataHandler SimpleCurveStartReceivingData
        {
            add { _psMultiCommSimple.SimpleCurveStartReceivingData += value; }
            remove { _psMultiCommSimple.SimpleCurveStartReceivingData -= value; }
        }

        /// <summary>
        /// Occurs when the devive's [state changed].
        /// </summary>
        public event MultiChannelStateChangedEventHandler StateChanged
        {
            add { _psMultiCommSimple.StateChanged += value; }
            remove { _psMultiCommSimple.StateChanged -= value; }
        }

        /// <summary>
        /// Occurs when a channel is [disconnected].
        /// </summary>
        public event MultiChannelDisconnectedEventHandler Disconnected
        {
            add { _psMultiCommSimple.Disconnected += value; }
            remove { _psMultiCommSimple.Disconnected -= value; }
        }
        #endregion

        public void Dispose()
        {
            _psMultiCommSimple.Dispose();
        }
    }
}
